--[[============================================================================
main.lua
============================================================================]]--

--[[
Grab sample keyzones 0.1

Grabs instrument keyzones with sample data from another instrument

-early. will have bugs.

-todo things for starters:
 -set min-max guards for several things
  -instrument no.
  -max no. of samples?
  -...

--]]




-- Placeholder for the dialog
local dialog = nil

-- Placeholder to expose the ViewBuilder outside the show_dialog() function
local vb = nil

-- Reload the script whenever this file is saved. 
-- Additionally, execute the attached function.
_AUTO_RELOAD_DEBUG = function()
  
end

-- Read from the manifest.xml file.
class "RenoiseScriptingTool" (renoise.Document.DocumentNode)
  function RenoiseScriptingTool:__init()    
    renoise.Document.DocumentNode.__init(self) 
    self:add_property("Name", "Untitled Tool")
    self:add_property("Id", "Unknown Id")
  end

local manifest = RenoiseScriptingTool()
local ok,err = manifest:load_from("manifest.xml")
local tool_name = manifest:property("Name").value
local tool_id = manifest:property("Id").value

local noteonlayer = renoise.Instrument.LAYER_NOTE_ON
local noteofflayer = renoise.Instrument.LAYER_NOTE_OFF

local gui_values
local song

--------------------------------------------------------------------------------
-- Main functions
--------------------------------------------------------------------------------

local function grab_keyzones()
    
  song = renoise.song()

  local instr_from_i = gui_values.instr_from_i.value + 1 --the +1 for lua / renoise mismatch
  local instr_from = song:instrument(instr_from_i)
  local instr_to = song.selected_instrument

  -----------------------------------set layers to grab
  local grab_layers = table.create()
  if gui_values.layer_on.value then
   grab_layers:insert(noteonlayer) 
  end
  if gui_values.layer_off.value then
   grab_layers:insert(noteofflayer) 
  end

  --------------------------------------------------layers loop-----------------

  for key, this_layer in pairs(grab_layers) do
    local kzs_from = #instr_from.sample_mappings[this_layer] --number of keyzonemappings in 'from' (note on layer)
    local kzs_to = #instr_to.sample_mappings[this_layer] -- in 'to' (note on layer)

    local samples_to = #instr_to.samples

    --note on layer

    local t_copied_samples = {}
    --[[
    this stores indexes of samples that get copied from the 'from' instrument
    this data is needed to map samples that have duplicate mappings in 'from' instrument

    data is stored as tables within table, formatted as follows
    t_copied_samples
      !
      1--{sample_index_from, sample_index_to}
      !
      2--{sample_index_from, sample_index_to}
      !
      ... et cetera

      sample_index_from = the original index of sample in the 'from'-instrument
      sample_index_to = the index of sample that is copied into 'to'-instrument

      --]]

    local stats_kzs = kzs_from
    local stats_spls = 0
    local stats_duplicates = 0

    for i = 1, kzs_from do

      --update sample count
      samples_to = #instr_to.samples

      local it_kz = i + kzs_to

      local mapping_from = instr_from:sample_mapping(this_layer, i)
      local mapping_to

      --'from sample' pointer
      local spl_i_from =  mapping_from.sample_index
      local spl_from = instr_from:sample(spl_i_from)
      local spl_i_to
      local spl_to


      --check if sample is already copied
      local is_copied = false --reset flag
      for ii = 1, #t_copied_samples do
        if t_copied_samples[ii][1] == spl_i_from then
          --remap
          spl_i_to = t_copied_samples[ii][2]
          --set flag
          is_copied = true
          break
        end
      end

      --check if the sample already exists in this instrument
      local is_duplicate = false --reset flag
      if not(is_copied) and (gui_values.smart_remap.value == true) then --smarty option

        for ii = 1, samples_to do
          --duplicate check
          --compare sample names (only... this is error prone.)
          local name_to_temp = instr_to:sample(ii).name

          if spl_from.name == name_to_temp then
            --remap
            spl_i_to = ii
            --set_flag
            is_duplicate = true
            --treat as copied
            table.insert(t_copied_samples, {spl_i_from, ii})
            --update stats
            stats_duplicates = stats_duplicates + 1
            break
          end

        end
      
      end

      local need_new = not(is_copied) and not(is_duplicate)

      if need_new == true then
        --need new sample slot
        spl_to = instr_to:insert_sample_at(samples_to + 1)
        spl_i_to = samples_to + 1
        --copy
        spl_to:copy_from(spl_from)
        --update t_copied_samples -table
        table.insert(t_copied_samples, {spl_i_from, spl_i_to})
        --update stats
        stats_spls = stats_spls + 1
      end

      --create new mapping, set values from 'from'-mapping
      mapping_to = instr_to:insert_sample_mapping(this_layer, spl_i_to, mapping_from.base_note, mapping_from.note_range, mapping_from.velocity_range)
      --copy yet nunset properties
      mapping_to.use_envelopes = mapping_from.use_envelopes
      mapping_to.map_velocity_to_volume = mapping_from.map_velocity_to_volume

    end
  renoise.app():show_status(string.format("Imported %i keyzones, %i samples (found %i duplicate samples)" , stats_kzs, stats_spls, stats_duplicates))
  end--layers loop


end


--------------------------------------------------------------------------------
-- GUI
--------------------------------------------------------------------------------

local function show_dialog()

  -- This block makes sure a non-modal dialog is shown once.
  -- If the dialog is already opened, it will be focused.
  if dialog and dialog.visible then
    dialog:show()
    return
  end
  
  if gui_values == nil then
    gui_values = renoise.Document.create("GrabSKzsGuiValues"){
      instr_from_i = 1,--for vb control binding
      layer_on = true,
      layer_off = true,
      smart_remap = true  --for vb control binding
    }
  end

  --set keyzone tab
  renoise.app().window.active_middle_frame = renoise.ApplicationWindow.MIDDLE_FRAME_KEYZONE_EDITOR

  -- The ViewBuilder is the basis
  vb = renoise.ViewBuilder()
  
  -- The content of the dialog, built with the ViewBuilder.
  local content = vb:column {
    margin = 10,
    vb:row {
      vb:text {
        text = "Grab from instrument "
      },
      vb:valuebox{
        -- bind to something
        bind = gui_values.instr_from_i,
        value = 0,
        tooltip = "Select instrument to grab keyzones from"
      }
    },

    vb:row {
      vb:text {
        text = "Smart remap "
      },
      vb:checkbox {
        bind = gui_values.smart_remap,
        tooltip = "Check for duplicate samples. (by name only)"
      }
    },
    vb:row {
      vb:text {
        text = "Note-on Layer"
      },
      vb:checkbox {
        bind = gui_values.layer_on,
        tooltip = "Get keyzones from note-on layer"
      }
    },
    vb:row {
      vb:text {
        text = "Note-off Layer"
      },
      vb:checkbox {
        bind = gui_values.layer_off,
        tooltip = "Get keyzones from note-off layer"
      }
    },
    --]]
    vb:button{
      text = "GRAB",
      notifier = grab_keyzones
    }
  } 
  
  dialog = renoise.app():show_custom_dialog(tool_name, content)  

end


--------------------------------------------------------------------------------
-- Menu entries
--------------------------------------------------------------------------------

renoise.tool():add_menu_entry {
  name = "Sample Mappings:"..tool_name.."...",
  invoke = show_dialog  
}


--------------------------------------------------------------------------------
-- Key Binding
--------------------------------------------------------------------------------

--[[
renoise.tool():add_keybinding {
  name = "Global:Tools:" .. tool_name.."...",
  invoke = show_dialog
}
--]]


--------------------------------------------------------------------------------
-- MIDI Mapping
--------------------------------------------------------------------------------

--[[
renoise.tool():add_midi_mapping {
  name = tool_id..":Show Dialog...",
  invoke = show_dialog
}
--]]
